@file:OptIn(LowLevelApi::class)
@file:JvmName("_Row2RowKt")
@file:Suppress("DuplicatedCode")

package ru.landgrafhomyak.db.serdha0.user_commons.executors

import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlin.jvm.JvmName
import ru.landgrafhomyak.db.serdha0.api.LowLevelApi
import ru.landgrafhomyak.db.serdha0.api.queries._Query
import ru.landgrafhomyak.db.serdha0.api.runtime.OutputRow
import ru.landgrafhomyak.db.serdha0.api.runtime.Transaction

public class ExpectedOneRowException : Error("Expected at least one row, but table is empty")
public class TooManyRowsException : Error("Expected exactly one row, but got more")

public suspend inline fun <qUE : Any, R> Transaction.selectExactlyOneOrError(
	compiledQuery: _Query.Params2Table<qUE>,
	params: RowProducer0<qUE>,
	result: (OutputRow<qUE>) -> R
): R {
	contract {
		callsInPlace(result, InvocationKind.EXACTLY_ONCE)
	}
	_safeAutoClose_IO(
		iRow = this._executeQuery(compiledQuery),
		inputAction = params::initializeRow,
		outputAction = { oRow ->
			if (!oRow._next())
				throw ExpectedOneRowException()
			_safeAutoClose(onSuccess = { if (oRow._next()) throw TooManyRowsException() }) {
				return result(oRow)
			}
		}
	)
}

public suspend inline fun <qUE : Any, R> Transaction.selectExactlyOneOrError(
	compiledQuery: _Query.Params2Table<qUE>,
	params: RowProducer0<qUE>,
	result: RowConsumer0<qUE, R>
): R = this.selectExactlyOneOrError(compiledQuery, params, result::transformRow)

public suspend inline fun <qUE : Any, R> Transaction.selectExactlyOneOrError(
	compiledQuery: _Query.Void2Table<qUE>,
	result: (OutputRow<qUE>) -> R
): R {
	contract {
		callsInPlace(result, InvocationKind.EXACTLY_ONCE)
	}
	_safeAutoClose_O(
		oRow = this._executeQuery(compiledQuery),
		outputAction = { oRow ->
			if (!oRow._next())
				throw ExpectedOneRowException()
			_safeAutoClose(onSuccess = { if (oRow._next()) throw TooManyRowsException() }) {
				return result(oRow)
			}
		}
	)
}

public suspend inline fun <qUE : Any, R> Transaction.selectExactlyOneOrError(
	compiledQuery: _Query.Void2Table<qUE>,
	result: RowConsumer0<qUE, R>
): R = this.selectExactlyOneOrError(compiledQuery, result::transformRow)

public suspend inline fun <qUE : Any, R> Transaction.selectExactlyOneOrNull(
	compiledQuery: _Query.Params2Table<qUE>,
	params: RowProducer0<qUE>,
	result: (OutputRow<qUE>) -> R
): R? {
	contract {
		callsInPlace(result, InvocationKind.AT_MOST_ONCE)
	}
	_safeAutoClose_IO(
		iRow = this._executeQuery(compiledQuery),
		inputAction = params::initializeRow,
		outputAction = { oRow ->
			if (!oRow._next())
				return null
			_safeAutoClose(onSuccess = { if (oRow._next()) throw TooManyRowsException() }) {
				return result(oRow)
			}
		}
	)
}

public suspend inline fun <qUE : Any, R> Transaction.selectExactlyOneOrNull(
	compiledQuery: _Query.Params2Table<qUE>,
	params: RowProducer0<qUE>,
	result: RowConsumer0<qUE, R>
): R? = this.selectExactlyOneOrNull(compiledQuery, params, result::transformRow)

public suspend inline fun <qUE : Any, R> Transaction.selectExactlyOneOrNull(
	compiledQuery: _Query.Void2Table<qUE>,
	result: (OutputRow<qUE>) -> R
): R? {
	contract {
		callsInPlace(result, InvocationKind.AT_MOST_ONCE)
	}
	_safeAutoClose_O(
		oRow = this._executeQuery(compiledQuery),
		outputAction = { oRow ->
			if (!oRow._next())
				return null
			_safeAutoClose(onSuccess = { if (oRow._next()) throw TooManyRowsException() }) {
				return result(oRow)
			}
		}
	)
}

public suspend inline fun <qUE : Any, R> Transaction.selectExactlyOneOrNull(
	compiledQuery: _Query.Void2Table<qUE>,
	result: RowConsumer0<qUE, R>
): R? = this.selectExactlyOneOrNull(compiledQuery, result::transformRow)

public suspend inline fun <qUE : Any, R> Transaction.selectFirstOrError(
	compiledQuery: _Query.Params2Table<qUE>,
	params: RowProducer0<qUE>,
	result: (OutputRow<qUE>) -> R
): R {
	contract {
		callsInPlace(result, InvocationKind.EXACTLY_ONCE)
	}
	_safeAutoClose_IO(
		iRow = this._executeQuery(compiledQuery),
		inputAction = params::initializeRow,
		outputAction = { oRow ->
			if (!oRow._next())
				throw ExpectedOneRowException()
			return result(oRow)
		}
	)
}

public suspend inline fun <qUE : Any, R> Transaction.selectFirstOrError(
	compiledQuery: _Query.Params2Table<qUE>,
	params: RowProducer0<qUE>,
	result: RowConsumer0<qUE, R>
): R = this.selectFirstOrError(compiledQuery, params, result::transformRow)

public suspend inline fun <qUE : Any, R> Transaction.selectFirstOrError(
	compiledQuery: _Query.Void2Table<qUE>,
	result: (OutputRow<qUE>) -> R
): R {
	contract {
		callsInPlace(result, InvocationKind.EXACTLY_ONCE)
	}
	_safeAutoClose_O(
		oRow = this._executeQuery(compiledQuery),
		outputAction = { oRow ->
			if (!oRow._next())
				throw ExpectedOneRowException()
			return result(oRow)
		}
	)
}

public suspend inline fun <qUE : Any, R> Transaction.selectFirstOrError(
	compiledQuery: _Query.Void2Table<qUE>,
	result: RowConsumer0<qUE, R>
): R = this.selectFirstOrError(compiledQuery, result::transformRow)

public suspend inline fun <qUE : Any, R> Transaction.selectFirstOrNull(
	compiledQuery: _Query.Params2Table<qUE>,
	params: RowProducer0<qUE>,
	result: (OutputRow<qUE>) -> R
): R? {
	contract {
		callsInPlace(result, InvocationKind.AT_MOST_ONCE)
	}
	_safeAutoClose_IO(
		iRow = this._executeQuery(compiledQuery),
		inputAction = params::initializeRow,
		outputAction = { oRow ->
			if (!oRow._next())
				return null
			return result(oRow)
		}
	)
}

public suspend inline fun <qUE : Any, R> Transaction.selectFirstOrNull(
	compiledQuery: _Query.Params2Table<qUE>,
	params: RowProducer0<qUE>,
	result: RowConsumer0<qUE, R>
): R? = this.selectFirstOrNull(compiledQuery, params, result::transformRow)

public suspend inline fun <qUE : Any, R> Transaction.selectFirstOrNull(
	compiledQuery: _Query.Void2Table<qUE>,
	result: (OutputRow<qUE>) -> R
): R? {
	contract {
		callsInPlace(result, InvocationKind.AT_MOST_ONCE)
	}
	_safeAutoClose_O(
		oRow = this._executeQuery(compiledQuery),
		outputAction = { oRow ->
			if (!oRow._next())
				return null
			return result(oRow)
		}
	)
}

public suspend inline fun <qUE : Any, R> Transaction.selectFirstOrNull(
	compiledQuery: _Query.Void2Table<qUE>,
	result: RowConsumer0<qUE, R>
): R? = this.selectFirstOrNull(compiledQuery, result::transformRow)